//
// Hardware.cpp:
//	Implementation of classes parallelPort and gloveDriver.
//
// Purpose:
//	Please see "hardware.hpp".
//
// Author(s): Mark Pflaging   C-serve  70233,1552
// Copyright 1992   Mark Thomas Pflaging
//
// Also see authorships in OOPGLOVE.DOC.
// Date:    5/15/92 - 6/4/92 (Release 1)
//	    8/9/92 - 9/21/92 (Release 2)
//
#include "hardware.hpp"
#include <stdlib.h>
#include <conio.h>

char * gloveDriver::title = "Timings";

parallelPort::parallelPort(InitFile & ini)
{
	char * altPort = ini.find("Parallel Port", "HexAlternatePort", "0x3BC");
	userOutPort = (int)(strtoul(altPort, NULL, 0));
}

parallelPort::parallelPort( tPort p )
{
	setPort(p);
}

parallelPort::setPort( tPort p )
{
	switch( p ) {
		case LPT1:
			set_glove_inPort( 0x379 );
			set_glove_outPort( 0x378 );
			break;
		case LPT2:
			set_glove_inPort( 0x279 );
			set_glove_outPort( 0x278 );
			break;
		case LPT3:
			set_glove_inPort( userOutPort + 1);
			set_glove_outPort( userOutPort );
			break;
		default:
			C0L0();
			cout << "Invalid port number." << endl;
			getch();
			exit(2);
	}
	return 0;
}

// count_once uses the precision timer to see how much time
// fdelay() is really using up.
int machineTimer::count_once()
{
	ptimer.reset();
	ptimer.start();
	fdelay();
	ptimer.stop();
	return ((int)ptimer.count());
}

// Following procedure averages 100 calls to count_once().
int machineTimer::get_count()
{
	long total = 0;
	for (int i = 0; i < 100; i ++)
		total += (long)count_once();
	int retval = (int)(total/(long)i);
	// round the result.
	if ((total - ((long)retval * (long)i)) >= (long)(i/2))
		retval ++;
	return retval;
}

void machineTimer::multipliedDelay(int multiplier)
{
	for (int i = 0; i < multiplier; i ++) fdelay();
}

// Some of the code in machineTimer::machineTimer() is
// courtesy Dream Park Research Company.
void machineTimer::machineTimer(unsigned int val, void (*func) (void),
	char * ascN, char * ascD) :
	timing_func(func), val(val), test_value(0)
{
	if ((ascN) && (ascD)) {
		Num = strtoul(ascN, NULL, 10);
		Den = strtoul(ascD, NULL, 10);
		return;
	}

	unsigned long int delta_D = 2;		// start delta at .5

	Num = 1;
	Den = 1;

	unsigned int mycount;
	// find over approximation for N
	while ((mycount = count_once()) < val)
		Num ++;

	// N/D guaranteed to be too high.
	unsigned int old_count = 0;
	int tolerance = val/100 + 1;
	for ( int j = 0 ; j < 10; ++j)  // find correct ratio between
	{                  // N and N-1
		mycount = get_count();
		// Are we within the tolerance?  Then why continue?
		if (abs(val - mycount) <= tolerance) {
			test_value = mycount;
			return;
		}
		unsigned long int oldN = Num;
		unsigned long int oldD = Den;
		if (mycount > val)	// fdelay is too slow
			Num = (Num * delta_D) - (Den);
		 else               //fdelay is too fast
			Num = (Num * delta_D) + (Den);


		 Den = Den * delta_D;
		 delta_D = delta_D * 2;

		//find lowest common denominator
		while ((((Num) &1) == 0) && (((Den) &1) == 0)) {
			Num = (Num)>>1;
			Den = (Den)>>1;
		}
		// Is the process degenerating????
		if (((old_count < val) && (mycount < old_count)) ||
		    ((old_count > val) && (mycount > old_count))) {
			// mycount is not really better!!
			// So restore old N and D.
			Num = oldN;
			Den = oldD;
			test_value = old_count;
			return;
		}
		else old_count = mycount;
	}
	test_value = mycount;
	return;
}

// so compiler won't optimize the dummy loop in fdelay().
// #pragma optimize ( ,{off})

// I substituted the following loops for fdelay() and then used
// the precision timer to time them.  Then I used them with the
// gloveDriver class to see which values would work.  This is
// how I determined the range for D2BITS and D2BYTES.
/**** D2BITS between 78 & 118 (33 MHz)
			for (int i2 = 0; i2 < 78 ; i2++)
				i2 = i2;
************/

/**** D2BYTES between 250 & 2500 (33 Mhz)
				for (int i2 = 0; i2 < 250; i2++)
					i2 = i2;
***************/

//void dumb_func()
//{
//	static int counter;
//	counter ++;
//	return;
//}

void machineTimer::fdelay(unsigned long N, unsigned long D,
	unsigned long val)
{
	long i;

	 i = ( long ) ( (N) * val );
	 for( ; i > 0; i -= (D) ) // dumb_func()
		;
}

void machineTimer::fdelay()
{
	long i;

	if (timing_func) timing_func();
	else {
		 i = ( long ) ( (Num) * val );
		 for( ; i > 0; i -= (Den) )
//			dumb_func()
			;
	}
}

void machineTimer::fdelay(int newval)
{
	unsigned long tempval = val;
	val = (unsigned long) newval;
	fdelay();
	val = tempval;
	return;
}

// > 15 us between reading in 2 bits
// > 85 us between reading in 2 bytes
gloveDriver::gloveDriver(InitFile & ini) : parallelPort(ini),
	segaISR(gloveISR),
	D2BITS(*(new machineTimer(
		ini.find(title, "D2BITS", 15), NULL,
		ini.find(title, "D2BITS.N"),
		ini.find(title, "D2BITS.D")
		))),
	D2BYTES(*(new machineTimer(
		ini.find(title, "D2BYTES", 100), NULL,
		ini.find(title, "D2BYTES.N"),
		ini.find(title, "D2BYTES.D")
		))),
	D2SLOW(*(new machineTimer(
		ini.find(title, "D2SLOW", 1000), NULL,
		ini.find(title, "D2SLOW.N"),
		ini.find(title, "D2SLOW.D")
		))),
	unready(0), glove_ignore(0), nmissed(0),
	glove_tc(
		ini.find(title, "MillisecsPerSample", 100))

{
	cout << "Waiting for glove...." << endl;
	glove_pause = glove_tc;
	tPort port = (tPort)'1';
	int done;
	do {
		setPort((tPort)(port));
		port = (tPort)(port + 1);
		if (port > LPT3) port = LPT1;
		int try = 0;
		do {
			// set PG into 'hires' mode
			D2SLOW.multipliedDelay(10);
			set_glove_hiResMode();
			// some time for the glove controller to relax
			D2SLOW.multipliedDelay(30);
			try ++;
			done = waitUntilReady();
		} while ((try < 1) && !done);
	} while (!done);
	forISR = this;
	run();
}

void gloveDriver::test()
{
	D2BITS.test();
	D2BYTES.test();
	D2SLOW.test();
}

gloveDriver::waitUntilReady()
{
	int i = 0, success;
	while( (!(success = is_glove_ready())) && (i < 30)) {
		D2SLOW.multipliedDelay(3);
		i ++;
	}
	if (!success)
		return 0;
	return 1;
}

// Call this function to check if glove data has been updated!
int gloveDriver::yield()
{
	if (nmissed) {
		disable();
		nmissed = 0;
		enable();
		return 1;    /* Interesting spot for a breakpoint. */
	}
	return 0;
}

void gloveDriver::waitForSample()
{
	unsigned long i = 0;
//	if (nmissed > 1) {
//		return;    /* Interesting spot for a breakpoint. */
//	}
	while (!yield()) { i ++; }
	return;
}

// gloveDriver::gloveHandler has to be a static function because
// it is called from an interrupt service routine.  But it
// needs access to non-static data.  So we need to get a
// pointer to 'this' into the scope of this function.  The
// only way to do it is to "cheat" - have a static variable
// that points to the one and only glove device driver.
// I was hoping to be able to have more than one gloveDriver
// active at one time, but it's going to be a little more
// difficult than I thought....

gloveDriver * gloveDriver::forISR;
#define MAXRESET  100	// failed reads before resetting glove

void gloveDriver::gloveISR()
{
	forISR->gloveHandler();
}

void gloveDriver::gloveHandler()
{
	int n;

	// wait out setup time if required
	if(glove_ignore) {
		glove_ignore--;
		return;
	}

	if ((--glove_pause) > 0) return;

	if (!is_glove_ready()) {
		if (++unready > MAXRESET) {
			// glove not responding... reset it
			unready = 0;
			set_glove_hiResMode();
		}
		glove_pause = 1;
	}
	else {
		// data ready!
		glove_pause = glove_tc;
		D2BYTES.fdelay();
		unready = 0;
		updateGloveData();

		// flag data as new
		++nmissed;
	}
}

unsigned char gloveDriver::get_glove_byte ()
{
	// read a byte from glove <rolled code>

	register int i;
	register unsigned char x = 0;

	C1L0 ();                       // generate a reset (latch) pulse
	C1L1 ();
	D2BITS.fdelay();
	C1L0 ();

	for( i = 0; i < 8; i++ )
	{
		x <<= 1;
		x += BITREAD();
		C0L0 ();
		C1L0 ();  // pulse
	}

	return( x );  // return the byte
}

int gloveDriver::is_glove_ready()
{
	// returns 1 if glove ready, 0 otherwise.

	int f = get_glove_byte();

	return( ( f== 0xA0 ) ? 1 : 0 );
}

int gloveDriver::updateGloveData()
{
	// read 6 byte data packet
	register unsigned char *bp;

	bp = getDataArea();

	*bp++ = get_glove_byte();    // read data
	D2BYTES.fdelay();
	*bp++ = get_glove_byte();
	D2BYTES.fdelay();
	*bp++ = get_glove_byte();
	D2BYTES.fdelay();
	*bp++ = get_glove_byte();
	D2BYTES.fdelay();
	*bp++ = get_glove_byte();
	D2BYTES.fdelay();
	*bp++ = get_glove_byte();
	D2BYTES.fdelay();
	// throwaways (speeds up polling later)
	// ???
	get_glove_byte();
	D2BYTES.fdelay();
	get_glove_byte();
	return 0;
}

/*  HIRES ENTRY CODES
byte:
1- any value between $05 and $31
2- only $C1 and $81 work OK
3- no effect
4- no effect
5- no effect
6- only $FF works
7- seems to affect read rate slightly, 1 fastest
   byte 7 = 0 causes LED's to stay on continuously, seems smoother. (mtp)
*/

static int hires_code[] = {
	0x06, 0xC1, 0x08, 0x00, 0x02, 0xFF, /* 0x01 */ 0x00
};


void gloveDriver::set_glove_hiResMode()
{
	// enter HIRES mode <rolled code- speed unimportant>.
	// Can somebody explain what he means by that?  -mtp

	int i, j, k;
			     // dummy read 4 bits from glove:
	C1L0 (); C1L1 ();    // generate a reset (latch) pulse
	D2BITS.fdelay();
	C1L0 ();

	D2BITS.fdelay();
	C0L0 (); C1L0 ();    // pulse clock
	D2BITS.fdelay();
	C0L0 (); C1L0 ();    // pulse clock
	D2BITS.fdelay();
	C0L0 (); C1L0 ();    // pulse clock
	D2BITS.fdelay();
	C0L0 (); C1L0 ();    // pulse clock

	// handshake for command code?
	C1L0 ();
	D2SLOW.multipliedDelay(7);	// 7212 us delay
	D2BYTES.multipliedDelay(2);
	C1L1 ();

	D2SLOW.multipliedDelay(2);	// 2260 us delay
	D2BYTES.multipliedDelay(3);

	for( i = 0; i < 7 ; i++ )       // send 7 bytes
	{
		k = hires_code[ i ];
		for (j = 0; j < 8; j ++)   // 8 bits per byte, MSB first
		{
			if( k & 0x80 ) {
				C1L1();
				C0L1();
				C1L1();
			}
			else {
				C1L0();
				C0L0();
				C1L0();
			}
			k <<= 1;
			D2BITS.fdelay();
		}
		D2BYTES.fdelay();
	}

	D2SLOW.fdelay();    // 892 us delay (end of 7. byte) */

	C1L0();          // drop the reset line
	// some time for the glove controller to relax
	glove_ignore = 100;

}

void machineTimer::test()
{
	cout << "N/D=" << int(Num) << "/" << int(Den) << " for ";
	cout << val << "usec delay.";
	if (test_value)
		cout << " (100 sample avg. = " << test_value
			<< " usec)";
	else cout << " (Precomputed).";
	cout << endl;
}
